import 'package:json_path/src/expression/bool_expression.dart';
import 'package:json_path/src/expression/expression.dart';
import 'package:json_path/src/expression/nodes_expression.dart';
import 'package:json_path/src/expression/value_expression.dart';
import 'package:json_path/src/fun/built_in/count_fun.dart';
import 'package:json_path/src/fun/built_in/length_fun.dart';
import 'package:json_path/src/fun/built_in/match_fun.dart';
import 'package:json_path/src/fun/fun.dart';
import 'package:json_path/src/fun/fun_call.dart';
import 'package:json_path/src/fun/fun_factory.dart';
import 'package:json_path/src/fun/search_fun.dart';
import 'package:json_path/src/grammar/array_index.dart';
import 'package:json_path/src/grammar/array_slice.dart';
import 'package:json_path/src/grammar/cmp_operator.dart';
import 'package:json_path/src/grammar/compare.dart';
import 'package:json_path/src/grammar/dot_name.dart';
import 'package:json_path/src/grammar/field_selector.dart';
import 'package:json_path/src/grammar/filter_selector.dart';
import 'package:json_path/src/grammar/fun_name.dart';
import 'package:json_path/src/grammar/literal.dart';
import 'package:json_path/src/grammar/node_selector.dart';
import 'package:json_path/src/grammar/parser_ext.dart';
import 'package:json_path/src/grammar/select_all_recuresively.dart';
import 'package:json_path/src/grammar/sequence_selector.dart';
import 'package:json_path/src/grammar/strings.dart';
import 'package:json_path/src/grammar/union_selector.dart';
import 'package:json_path/src/grammar/wildcard.dart';
import 'package:maybe_just_nothing/maybe_just_nothing.dart';
import 'package:petitparser/petitparser.dart';

class JsonPathGrammarDefinition extends GrammarDefinition<NodesExpression> {
  JsonPathGrammarDefinition(Iterable<Fun> userFunctions)
      : _fun = FunFactory(userFunctions.followedBy(_builtInFunctions));

  static const _builtInFunctions = <Fun>[
    LengthFun(),
    CountFun(),
    MatchFun(),
    SearchFun(),
  ];

  final FunFactory _fun;

  @override
  Parser<NodesExpression> start() => ref0(_absPath).end();

  Parser<NodeSelector> _unionElement() => [
        arraySlice,
        arrayIndex,
        wildcard,
        quotedString.map(fieldSelector),
        ref0(_expressionFilter)
      ].toChoiceParser().trim();

  Parser<NodeSelector> _nextUnionElement() =>
      ref0(_unionElement).skip(before: char(','));

  Parser<Iterable<NodeSelector>> _unionContent() => [
        ref0(_unionElement).map((v) => [v]),
        ref0(_nextUnionElement).star()
      ].toSequenceParser().map((v) => v.expand((e) => e));

  Parser<NodeSelector> _union() => ref0(_unionContent)
      .skip(before: char('['), after: char(']'))
      .map(unionSelector);

  Parser<NodeSelector> _recursion() => [
        wildcard,
        ref0(_union),
        memberNameShorthand,
      ]
          .toChoiceParser()
          .skip(before: string('..'))
          .map((value) => sequenceSelector([selectAllRecursively, value]));

  Parser<BoolExpression> _parenExpr() => ref1(
        _negatable,
        ref0(_logicalExpr)
            .skip(before: char('(').trim(), after: char(')').trim()),
      );

  Parser<BoolExpression> _negation(Parser<BoolExpression> p) =>
      p.skip(before: char('!').trim()).map((expr) => expr.map((v) => !v));

  Parser<Expression> _funArgument() => [
        literal,
        ref0(_filterPath),
      ].toChoiceParser().trim();

  Parser<Expression> _funNextArgument() =>
      _funArgument().skip(before: char(',').trim()).trim();

  Parser<List<Expression>> _functionArguments() => [
        _funArgument().map((v) => [v]),
        _funNextArgument().star()
      ].toSequenceParser().map((v) => v.expand((e) => e).toList());

  Parser<T> _funCall<T>(T? Function(FunCall) funMaker) =>
      (funName & _functionArguments().skip(before: char('('), after: char(')')))
          .map((v) => FunCall(v[0], v[1]))
          .tryMap((call) =>
              funMaker(call) ??
              (throw Exception('No implementation for $call found')));

  Parser<ValueExpression> _comparableFunExpr() =>
      ref1(_funCall, _fun.comparable);

  Parser<BoolExpression> _logicalFunExpr() => ref1(_funCall, _fun.logical);

  Parser<ValueExpression> _comparable() => [
        literal,
        ref0(_filterPath).map((it) => it.asValueExpression),
        ref0(_comparableFunExpr),
      ].toChoiceParser();

  Parser<BoolExpression> _cmpExpr() =>
      (ref0(_comparable) & cmpOperator & ref0(_comparable)).map((v) {
        final Expression<Maybe> left = v[0];
        final String op = v[1];
        final Expression<Maybe> right = v[2];

        return left.merge(right, (l, r) => compare(op, l, r));
      });

  Parser<BoolExpression> _logicalExpr() => ref0(_logicalOrExpr);

  Parser<BoolExpression> _logicalOrExpr() => [
        ref0(_logicalAndExpr).map((v) => [v]),
        ref0(_logicalAndExpr).skip(before: string('||').trim()).star(),
      ]
          .toSequenceParser()
          .map((v) => v.expand((e) => e))
          .map((tests) => tests.reduce((a, b) => a.merge(b, (a, b) => a || b)));

  Parser<BoolExpression> _logicalAndExpr() => [
        ref0(_basicExpr).map((v) => [v]),
        ref0(_basicExpr).skip(before: string('&&').trim()).star(),
      ]
          .toSequenceParser()
          .map((v) => v.expand((e) => e))
          .map((tests) => tests.reduce((a, b) => a.merge(b, (a, b) => a && b)));

  Parser<BoolExpression> _negatable(Parser<BoolExpression> p) =>
      [ref1(_negation, p), p].toChoiceParser();

  Parser<BoolExpression> _basicExpr() => [
        ref0(_parenExpr),
        ref0(_cmpExpr),
        ref0(_testExpr),
      ].toChoiceParser();

  Parser<NodesExpression> _filterPath() => [
        ref0(_relPath),
        ref0(_absPath),
      ].toChoiceParser();

  Parser<BoolExpression> _existenceTest() =>
      ref0(_filterPath).map((value) => value.asLogicalExpression);

  Parser<BoolExpression> _testExpr() => ref1(
      _negatable,
      [
        ref0(_existenceTest),
        ref0(_logicalFunExpr),
      ].toChoiceParser());

  Parser<NodeSelector> _expressionFilter() =>
      ref0(_logicalExpr).skip(before: string('?')).map(filterSelector);

  Parser<NodeSelector> _segment() => [
        dotName,
        ref0(_union),
        ref0(_recursion),
      ].toChoiceParser();

  Parser<NodesExpression> _segmentSequence() =>
      ref0(_segment).star().map(sequenceSelector).map(Expression.new);

  Parser<NodesExpression> _absPath() => ref0(_segmentSequence)
      .skip(before: char(r'$'))
      .map((expr) => Expression((node) => expr.call(node.root)));

  Parser<NodesExpression> _relPath() =>
      ref0(_segmentSequence).skip(before: char('@'));
}
